/*
 * $Id: pam_wheel.c,v 1.4 2001/01/14 23:33:48 udp Exp bsimpson $
 *
 * pam_wheel.c -- PAM module to check if user is a member of the 'wheel'
 *                group, when trying to become uid 0. for use with the
 *                Solaris 'su' command to emulate BSD semantics.
 *
 * setting 'forbidall' means only permit su to any account if in wheel.
 * default is deny all if named wheel group not found in system.
 *
 * Main change: create the audit trail correctly in syslog according to
 * severity/nature of log message (LOG_AUTH with LOG_WARN/LOG_DEBUG etc.)
 * 
 * User Datagram Protocol <udp@low-level.net> -- 24/12/2000
 *
 */
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>
#include <security/pam_appl.h>

#define WHEELGROUP "wheel"
#define ROOT_UID ((uid_t)0L)

extern int pam_sm_authenticate(
	pam_handle_t	 *pamh, 
	int		  flags,	
	int		  argc,
	char		**argv
)
{
	struct	passwd	  pwd;
	struct	group	  grp;
	char 		 *pwbuf = NULL;
	char 		 *grbuf = NULL;
	static const int  pwbuflen = 512;
	static const int  grbuflen = 512;
	char		 *this_user;
	char		 *desired_user;
	char		 *service;
	char		**mem;
	int 		  r;
	int		  i;
	int		  debug = 0;
	int		  forbidall = 0;

	/* Parse module options. */
	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i],"debug") == 0)
			debug++;
		if (strcmp(argv[i],"forbidall") == 0)
			forbidall++;
	}

	/* Get PAM environment: who are we trying to su to? */
        if ((r = pam_get_item(pamh, PAM_SERVICE, (void **)&service))
	   != PAM_SUCCESS ||
            (r = pam_get_user(pamh, &desired_user, NULL)) != PAM_SUCCESS)
                return r;

	if (debug) syslog(LOG_DEBUG,"wheel pam_sm_authenticate(%s %s), " \
	"flags = %x", service, desired_user?desired_user:"no-user", flags);

	if ((pwbuf = malloc(pwbuflen)) == NULL)
		return PAM_BUF_ERR;

	/* setting forbidall forbids su to all accounts if not in wheel */
	if (!forbidall) {
		/* Check if target account is uid 0 and twist accordingly */ 
		if (getpwnam_r(desired_user, &pwd, pwbuf, pwbuflen) == NULL) {
			r =  PAM_AUTH_ERR; goto cleanup;
		}
		if (pwd.pw_uid != ROOT_UID) {
			r = PAM_SUCCESS; goto cleanup;
		}
	}

	if (getpwuid_r(getuid(), &pwd, pwbuf, pwbuflen) == NULL) {
		r =  PAM_AUTH_ERR; goto cleanup;
	}
	this_user = pwd.pw_name; 

	if ((grbuf = malloc(grbuflen)) == NULL)
		return PAM_BUF_ERR;

	/* Check if user in wheel group */
	if ( (getgrnam_r(WHEELGROUP, &grp, grbuf, grbuflen) == NULL)
	     || ((mem = grp.gr_mem) == NULL) )
	{
		syslog(LOG_ERR, "wheel group does not exist, " \
		"or could not perform group list lookup");
		r = PAM_AUTH_ERR; goto cleanup;
	}
	while (*mem) {
		if (strcasecmp(this_user,*mem) == 0) {
			syslog(LOG_NOTICE, "user '%s' is in wheel group", 
				this_user);
			r =  PAM_SUCCESS; goto cleanup;
		}
		mem++;
	}

	syslog(LOG_WARNING, "user '%s' is NOT in wheel group", this_user);
	r = PAM_AUTH_ERR;
cleanup:
	if (grbuf) free(grbuf);
	if (pwbuf) free(pwbuf);
	return r;
}

/********************************************************************/

/* stubs */

extern int pam_sm_setcred(
        pam_handle_t    *pamh,
        int     flags,
        int     argc,
        const char      **argv)
{
	return PAM_SUCCESS;
}

extern int pam_sm_acct_mgmt(
        pam_handle_t    *pamh,
        int     flags,
        int     argc,
        const char      **argv)
{
	return PAM_SUCCESS;
}

extern int pam_sm_open_session(
        pam_handle_t    *pamh,
        int     flags,
        int     argc,
        const char      **argv)
{
	return PAM_SUCCESS;
}

extern int pam_sm_close_session(
        pam_handle_t    *pamh,
        int     flags,
        int     argc,
        const char      **argv)
{
	return PAM_SUCCESS;
}

extern int
pam_sm_chauthtok(
        pam_handle_t    *pamh,
        int             flags,
        int             argc,
        const char      **argv)
{
	return PAM_SUCCESS;
}

/********************************************************************/
